// QImageProperties and QImagePropertiesDlg
//
// Classes to work with GDI+ Image properties.
// Use at your own risk. Comments welcome.
//
// Version 1.0 (c) 2003, 10/20/2003, Sjaak Priester, Amsterdam.
// mailto:sjaak@sjaakpriester.nl

// Note: next GUIDs are defined in GdiplusImaging.h.
// However, in one of the versions of this file (dated 08/15/2002)
// there is an obvious bug: EncoderTransform is not defined, in stead
// EncoderQuality is defined two times, sure enough the second time
// with EncoderTransformation's guid-values.
// Strangely, an earlier version of GdiplusImaging.h (a 'prerelease' of 04/12/2001)
// has it right.
// I did not see any workaround but to manually change the offending line
// in the header.
// It's the second DEFINE_GUID(EncoderQuality... on line 59 of GdiplusImaging.h, where
// the guid data start with 0x8d0eb2d1. EncoderQuality should read EncoderTransformation.

#include "stdafx.h"
#include "QImageProperties.h"
#include "QImagePropertiesRes.h"

#if _MFC_VER < 0x0700
#include <atlconv.h>	// needed for SetProperty(PROPID, CString)
#endif

#define EncoderColorSpace	GUID_NULL		// documented, not in header (GDI+ 1.0)
#define EncoderImageItems	GUID_NULL
#define EncoderSaveAsCMYK	GUID_NULL

// A few property tags not (yet) defined in GdiplusImaging.h
// Source: www.exif.org, EXIF 2.2 specification
#define QPropertyTagSubjectArea					0x9214
#define QPropertyTagSubSecTime					0x9220
#define QPropertyTagSubSecTimeOriginal			0x9221
#define QPropertyTagCustomRendered				0xa401
#define QPropertyTagExposureMode				0xa402
#define QPropertyTagWhiteBalance				0xa403
#define QPropertyTagDigitalZoomRatio			0xa404
#define QPropertyTagFocalLengthIn35mmFilm		0xa405
#define QPropertyTagSceneCaptureType			0xa406
#define QPropertyTagGainControl					0xa407
#define QPropertyTagContrast					0xa408
#define QPropertyTagSaturation					0xa409
#define QPropertyTagSharpness					0xa40a
#define QPropertyTagDeviceSettingDescription	0xa40b
#define QPropertyTagSubjectDistanceRange		0xa40c
#define QPropertyTagImageUniqueID				0xa420

#define IDS_UNKNOWN		QDS_PIXELFORMATS

//======================
// utility functions

// From VC++ 7.0 on, CString has two very handy member functions AppendFormat().
// Alas, VC++ 6.0 hasn't. In stead we use these global functions.
void __cdecl CStringAppendFormat(CString& s, LPCTSTR sFormat, ...)
{
	va_list argList;
	va_start(argList, sFormat);

	CString sFormatted;
	sFormatted.FormatV(sFormat, argList);
	s += sFormatted;

	va_end( argList );
}

void __cdecl CStringAppendFormat(CString& s, UINT nFormatID, ...)
{
	va_list argList;
	va_start(argList, nFormatID);

	CString sFormat;
	VERIFY(sFormat.LoadString(nFormatID));
	// If you get an ASSERT here, chances are that you didn't include the
	// QImageProperties resources. These are in QImagePoperties.rc.
	// This file should be included in the application's main .rc-file.

	CString sFormatted;
	sFormatted.FormatV(sFormat, argList);
	s += sFormatted;

	va_end( argList );
}

//======================
// QImageProperties

QImageProperties::QImageProperties(Image * pImage)
: m_pImage(pImage)
{
}

QImageProperties::~QImageProperties(void)
{
}

// Caller must delete returned bytes
BYTE * QImageProperties::GetPropertyBytes(PROPID tag) const
{
	ASSERT(m_pImage);

	UINT nBytes = m_pImage->GetPropertyItemSize(tag);
	if (nBytes == 0) return NULL;

	BYTE * pBytes = new BYTE[nBytes];
	if (pBytes)
	{
		PropertyItem * pProp = (PropertyItem *) pBytes;
		if (m_pImage->GetPropertyItem(tag, nBytes, pProp) != Ok)
		{
			delete[] pBytes;
			return NULL;
		}
	}
	return pBytes;
}

bool QImageProperties::GetProperty(PROPID tag, CString& s) const
{
	BYTE * pBytes = GetPropertyBytes(tag);

	bool r = false;
	if (pBytes)
	{
		PropertyItem * pProp = (PropertyItem *) pBytes;
		if (pProp->type == PropertyTagTypeASCII)
		{
			s = (char *) pProp->value;
			r = true;
		}
	}
	delete[] pBytes;
	return r;
}

bool QImageProperties::GetProperty(PROPID tag, REAL& d) const
{
	BYTE * pBytes = GetPropertyBytes(tag);

	bool r = false;
	if (pBytes)
	{
		PropertyItem * pProp = (PropertyItem *) pBytes;
		if ((pProp->type == PropertyTagTypeRational || pProp->type == PropertyTagTypeSRational)
			&& pProp->length == 2 * sizeof(long))
		{
			double num, denom;
			if (pProp->type == PropertyTagTypeRational)
			{
				UINT * p = (UINT *) pProp->value;
				num = * p++;
				denom = * p;
			}
			else
			{
				int * p = (int *) pProp->value;
				num = * p++;
				denom = * p;
			}

			d = (float) (num / denom);
			r = true;
		}
	}
	delete[] pBytes;
	return r;
}

bool QImageProperties::SetProperty(PROPID tag, CString s)
{
	USES_CONVERSION;

	PropertyItem prop;
	prop.type = PropertyTagTypeASCII;
	prop.id = tag;
	prop.length = s.GetLength() + 1;
	char * p = T2A((LPTSTR)(LPCTSTR) s);
	prop.value = p;
	return (m_pImage->SetPropertyItem(& prop) == Ok);
}

CString QImageProperties::PropertyString(PROPID tag, LPCTSTR colon) const
{
	CString s = PropertyName(tag);
	s += colon;

	BYTE * pBytes = GetPropertyBytes(tag);
	if (pBytes)
	{
		PropertyItem * pProp = (PropertyItem *) pBytes;

		UINT base = 0;
		UINT max = 0;

		switch (pProp->id)
		{
		case PropertyTagResolutionUnit:
		case PropertyTagExifFocalResUnit:
		case PropertyTagThumbnailResolutionUnit:
			base = QDS_RESOLUTIONUNIT;
			max = 3;
			break;
		case PropertyTagCompression:
		case PropertyTagThumbnailCompression:
			base = QDS_COMPRESSION;
			max = 6;
			break;
		case PropertyTagPhotometricInterp:
			base = QDS_PHOTOMETRICINT;
			max = 6;
			break;
		case PropertyTagOrientation:
			base = QDS_ORIENTATION;
			max = 8;
			break;
		case PropertyTagExifExposureProg:
			base = QDS_EXPOSUREPROG;
			max = 8;
			break;
		case PropertyTagExifMeteringMode:
			base = QDS_METERINGMODE;
			max = 6;
			break;
		case PropertyTagExifLightSource:
			base = QDS_LIGHTSOURCE;
			max = 24;
			break;
		default:
			break;
		}

		if (base != 0)
		{
			BYTE * p = (BYTE *) pProp->value;
			BYTE d = * p;
			if (d >= 1 && d <= max)
			{
				CString sValue;
				VERIFY(sValue.LoadString(base + d));
				// If you get an ASSERT here, chances are that you didn't include the
				// QImageProperties resources. These are in QImagePoperties.rc.
				// This file should be included in the application's main .rc-file.

				s += sValue;
			}
			delete[] pBytes;
			return s;
		}

		UINT n = 0;

		UINT tp = pProp->type;
		if (tp > PropertyTagTypeSRational) tp = PropertyTagTypeUndefined;

		switch (tp)
		{
		case PropertyTagTypeASCII:
			s += (char *) pProp->value;
			delete[] pBytes;
			return s;

		case PropertyTagTypeShort:
			n = pProp->length / sizeof(short);
			break;
		case PropertyTagTypeLong:
		case PropertyTagTypeSLONG:
			n = pProp->length / sizeof(int);
			break;
		case PropertyTagTypeRational:
		case PropertyTagTypeSRational:
			n = pProp->length / (2 * sizeof(int));
			break;
//		case PropertyTagTypeByte:
//		case PropertyTagTypeUndefined:
		default:
			n = pProp->length / sizeof(BYTE);
			break;
		}

		CString sProp;

		if (n != 1)
		{
			if ((tp == PropertyTagTypeByte || tp == PropertyTagTypeUndefined) && n <= 4)
			{
				BYTE * p = (BYTE *) pProp->value;
				char c[5];
				char * pChars = c;
				bool bChars = true;
				for (UINT i = 0; i < n; i++)
				{
					CString sByte;
					sByte.Format(_T("%#02x "), * p);
					sProp += sByte;
					if (* p >= ' ') * pChars++ = * p;
					else bChars = false;
					p++;
				}

				if (bChars)
				{
					* pChars = _T('\0');
					sProp = c;
				}
			}
			else
			{
				CString sTagType;
				VERIFY(sTagType.LoadString(QDS_TAGTYPES + tp));
				// If you get an ASSERT here, chances are that you didn't include the
				// QImageProperties resources. These are in QImagePoperties.rc.
				// This file should be included in the application's main .rc-file.

				sProp.Format(_T("(%u %s)"), n, sTagType);
			}
		}
		else switch (tp)
		{
		case PropertyTagTypeShort:
			{
				short * p = (short *) pProp->value;
				UINT d = * p;
				sProp.Format(_T("%u (= %#04x)"), d, d);
			}
			break;
		case PropertyTagTypeLong:
			{
				UINT * p = (UINT *) pProp->value;
				sProp.Format(_T("%u (= %#08x)"), * p, * p);
			}
			break;
		case PropertyTagTypeRational:
			{
				UINT * p = (UINT *) pProp->value;
				UINT num = * p++;
				UINT denom = * p;
				if (denom)
					sProp.Format(_T("%s (%u/%u)"), RealToString((double) num/(double) denom),
						num, denom);
			}
			break;
		case PropertyTagTypeSLONG:
			{
				int * p = (int *) pProp->value;
				sProp.Format(_T("%d (= %#08x)"), * p, * p);
			}
			break;
		case PropertyTagTypeSRational:
			{
				int * p = (int *) pProp->value;
				int num = * p++;
				int denom = * p;
				if (denom)
					sProp.Format(_T("%s (%d/%d)"), RealToString((double) num/(double) denom),
						num, denom);
			}
			break;
//			case PropertyTagTypeByte:
//			case PropertyTagTypeUndefined:
		default:
			{
				BYTE * p = (BYTE *) pProp->value;
				UINT d = * p;
				sProp.Format(_T("%u (= %#02x)"), d, d);
			}
			break;
		}
		s += sProp;
		delete[] pBytes;
	}
	return s;
}

CString QImageProperties::AllProperties(PROPID * pSkipList, LPCTSTR colon) const
{
	LPCTSTR endl = _T("\r\n");
	LPCTSTR mult = _T(" \xd7 ");

	CString sProps;

	VERIFY(sProps.LoadString(QDS_FORMAT));
	// If you get an ASSERT here, chances are that you didn't include the
	// QImageProperties resources. These are in QImagePoperties.rc.
	// This file should be included in the application's main .rc-file.

	sProps += colon;
	GUID guid;
	if (m_pImage->GetRawFormat(& guid) == Ok) sProps += GuidString(guid);
	sProps += endl;

	::CStringAppendFormat(sProps, QDS_PIXELFORMAT);
	sProps += colon;
	PixelFormat pf = m_pImage->GetPixelFormat();
	UINT iPF = (pf & 0x000f) - 1;
	if (iPF == 15) iPF = 0;
	::CStringAppendFormat(sProps, QDS_PIXELFORMATS + iPF);
	sProps += endl;

	::CStringAppendFormat(sProps, QDS_PIXELS);
	sProps += colon;
	::CStringAppendFormat(sProps, _T("%d"), m_pImage->GetWidth());
	sProps += mult;
	::CStringAppendFormat(sProps, _T("%d"), m_pImage->GetHeight());
	sProps += endl;

	SizeF physSize;
	m_pImage->GetPhysicalDimension(& physSize);
	::CStringAppendFormat(sProps, QDS_PHYSDIMENSION);
	sProps += colon;
	sProps += RealToString(physSize.Width);
	sProps += mult;
	sProps += RealToString(physSize.Height);
	sProps += endl;

	::CStringAppendFormat(sProps, QDS_RESOLUTION);
	sProps += colon;
	sProps += RealToString(m_pImage->GetHorizontalResolution());
	sProps += mult;
	sProps += RealToString(m_pImage->GetVerticalResolution());
	sProps += endl;

	UINT flags = m_pImage->GetFlags();
	::CStringAppendFormat(sProps, QDS_FLAGS);
	sProps += colon;
	if (flags == 0) ::CStringAppendFormat(sProps, QDS_NONE);
	sProps += endl;
	UINT f = 0x0001;
	for (UINT i = 0; i < 13; i++)
	{
		if (f & flags)
		{
			sProps += _T("\t");
			::CStringAppendFormat(sProps, QDS_IMAGEFLAGS + i);
			sProps += endl;
		}
		f <<= 1;
	}

	UINT n = m_pImage->GetPropertyCount();
	if (n)
	{
		PROPID * pProps = new PROPID[n];
		if (m_pImage->GetPropertyIdList(n, pProps) == Ok)
			for (UINT i = 0; i < n; i++)
			{
				PROPID tag = pProps[i];
				if (pSkipList)
				{
					bool bSkip = false;

					PROPID * p = pSkipList;
					while (* p)
					{
						if (tag == * p)
						{
							bSkip = true;
							break;
						}
						++p;
					}
					if (bSkip) continue;
				}

				sProps += PropertyString(tag, colon) + endl;
			}

		delete[] pProps;
	}

	return sProps;
}

void QImageProperties::Copy(const QImageProperties& source,
							PROPID * pSkipList, bool bSkipThumbnail)
{
	if (source.m_pImage->GetLastStatus() != Ok) return;

	UINT nProps = source.m_pImage->GetPropertyCount();
	if (nProps)
	{
		PROPID * pPropList = new PROPID[nProps];
		source.m_pImage->GetPropertyIdList(nProps, pPropList);

		for (UINT i = 0; i < nProps; i++)
		{
			PROPID tag = pPropList[i];

			if (bSkipThumbnail
				&& tag >= PropertyTagThumbnailFormat
				&& tag <= PropertyTagChrominanceTable) continue;

			if (pSkipList)
			{
				bool bSkip = false;

				PROPID * p = pSkipList;
				while (* p)
				{
					if (tag == * p)
					{
						bSkip = true;
						break;
					}
					++p;
				}
				if (bSkip) continue;
			}

			BYTE * pBytes = source.GetPropertyBytes(pPropList[i]);

			if (pBytes)
			{
				PropertyItem * pProp = (PropertyItem *) pBytes;
				m_pImage->SetPropertyItem(pProp);
				delete[] pBytes;
			}
		}
		delete[] pPropList;
	}
}

#if _MSC_VER >= 1300

bool QImageProperties::SetProperty(PROPID tag, REAL d)
{
	if (d >= 0.0f) return SetProperty(tag, (UINT)(10000.0f * d), (UINT)10000);
	else return SetProperty(tag, (int)(10000.0f * d), (int)10000);
}

#endif

//=================
// static methods

/* static */
CString QImageProperties::GuidString(GUID& guid)
{
	const GUID * Guids[] =
	{
		// Formats
		&ImageFormatUndefined,
		&ImageFormatMemoryBMP,
		&ImageFormatBMP,
		&ImageFormatEMF,
		&ImageFormatWMF,
		&ImageFormatJPEG,
		&ImageFormatPNG,
		&ImageFormatGIF,
		&ImageFormatTIFF,
		&ImageFormatEXIF,
		&ImageFormatIcon,

		// Frame dimensions
		&FrameDimensionTime,
		&FrameDimensionResolution,
		&FrameDimensionPage,

		// Property sets
		&FormatIDImageInformation,
		&FormatIDJpegAppHeaders,

		// Encoder parameters
		&EncoderChrominanceTable,
		&EncoderColorDepth,
		&EncoderColorSpace,
		&EncoderCompression,
		&EncoderImageItems,
		&EncoderLuminanceTable,
		&EncoderQuality,
		&EncoderRenderMethod,
		&EncoderSaveFlag,
		&EncoderScanMethod,
		&EncoderVersion,
		&EncoderSaveAsCMYK,
		&EncoderTransformation
	};

	const int n = sizeof(Guids)/sizeof(GUID *);

	int id = IDS_UNKNOWN;
	for (int i = 0; i < n; i++)
		if (guid == * Guids[i])
		{
			id = QDS_GUID + i;
			break;
		}

	CString s;
	VERIFY(s.LoadString(id));

	return s;
}

/* static */
CString QImageProperties::PropertyName(PROPID tag)
{
	PROPID idPropTags[] =
	{
		PropertyTagExifIFD,
		PropertyTagGpsIFD,
		PropertyTagNewSubfileType,
		PropertyTagSubfileType,
		PropertyTagImageWidth,
		PropertyTagImageHeight,
		PropertyTagBitsPerSample,
		PropertyTagCompression,
		PropertyTagPhotometricInterp,
		PropertyTagThreshHolding,
		PropertyTagCellWidth,
		PropertyTagCellHeight,
		PropertyTagFillOrder,
		PropertyTagDocumentName,
		PropertyTagImageDescription,
		PropertyTagEquipMake,
		PropertyTagEquipModel,
		PropertyTagStripOffsets,
		PropertyTagOrientation,
		PropertyTagSamplesPerPixel,
		PropertyTagRowsPerStrip,
		PropertyTagStripBytesCount,
		PropertyTagMinSampleValue,
		PropertyTagMaxSampleValue,
		PropertyTagXResolution,
		PropertyTagYResolution,
		PropertyTagPlanarConfig,
		PropertyTagPageName,
		PropertyTagXPosition,
		PropertyTagYPosition,
		PropertyTagFreeOffset,
		PropertyTagFreeByteCounts,
		PropertyTagGrayResponseUnit,
		PropertyTagGrayResponseCurve,
		PropertyTagT4Option,
		PropertyTagT6Option,
		PropertyTagResolutionUnit,
		PropertyTagPageNumber,
		PropertyTagTransferFuncition,	// note: this spelling error is by MS
		PropertyTagSoftwareUsed,
		PropertyTagDateTime,
		PropertyTagArtist,
		PropertyTagHostComputer,
		PropertyTagPredictor,
		PropertyTagWhitePoint,
		PropertyTagPrimaryChromaticities,
		PropertyTagColorMap,
		PropertyTagHalftoneHints,
		PropertyTagTileWidth,
		PropertyTagTileLength,
		PropertyTagTileOffset,
		PropertyTagTileByteCounts,
		PropertyTagInkSet,
		PropertyTagInkNames,
		PropertyTagNumberOfInks,
		PropertyTagDotRange,
		PropertyTagTargetPrinter,
		PropertyTagExtraSamples,
		PropertyTagSampleFormat,
		PropertyTagSMinSampleValue,
		PropertyTagSMaxSampleValue,
		PropertyTagTransferRange,
		PropertyTagJPEGProc,
		PropertyTagJPEGInterFormat,
		PropertyTagJPEGInterLength,
		PropertyTagJPEGRestartInterval,
		PropertyTagJPEGLosslessPredictors,
		PropertyTagJPEGPointTransforms,
		PropertyTagJPEGQTables,
		PropertyTagJPEGDCTables,
		PropertyTagJPEGACTables,
		PropertyTagYCbCrCoefficients,
		PropertyTagYCbCrSubsampling,
		PropertyTagYCbCrPositioning,
		PropertyTagREFBlackWhite,
		PropertyTagICCProfile,
		PropertyTagGamma,
		PropertyTagICCProfileDescriptor,
		PropertyTagSRGBRenderingIntent,
		PropertyTagImageTitle,
		PropertyTagCopyright,
		PropertyTagResolutionXUnit,
		PropertyTagResolutionYUnit,
		PropertyTagResolutionXLengthUnit,
		PropertyTagResolutionYLengthUnit,
		PropertyTagPrintFlags,
		PropertyTagPrintFlagsVersion,
		PropertyTagPrintFlagsCrop,
		PropertyTagPrintFlagsBleedWidth,
		PropertyTagPrintFlagsBleedWidthScale,
		PropertyTagHalftoneLPI,
		PropertyTagHalftoneLPIUnit,
		PropertyTagHalftoneDegree,
		PropertyTagHalftoneShape,
		PropertyTagHalftoneMisc,
		PropertyTagHalftoneScreen,
		PropertyTagJPEGQuality,
		PropertyTagGridSize,
		PropertyTagThumbnailFormat,
		PropertyTagThumbnailWidth,
		PropertyTagThumbnailHeight,
		PropertyTagThumbnailColorDepth,
		PropertyTagThumbnailPlanes,
		PropertyTagThumbnailRawBytes,
		PropertyTagThumbnailSize,
		PropertyTagThumbnailCompressedSize,
		PropertyTagColorTransferFunction,
		PropertyTagThumbnailData,
		PropertyTagThumbnailImageWidth,
		PropertyTagThumbnailImageHeight,
		PropertyTagThumbnailBitsPerSample,
		PropertyTagThumbnailCompression,
		PropertyTagThumbnailPhotometricInterp,
		PropertyTagThumbnailImageDescription,
		PropertyTagThumbnailEquipMake,
		PropertyTagThumbnailEquipModel,
		PropertyTagThumbnailStripOffsets,
		PropertyTagThumbnailOrientation,
		PropertyTagThumbnailSamplesPerPixel,
		PropertyTagThumbnailRowsPerStrip,
		PropertyTagThumbnailStripBytesCount,
		PropertyTagThumbnailResolutionX,
		PropertyTagThumbnailResolutionY,
		PropertyTagThumbnailPlanarConfig,
		PropertyTagThumbnailResolutionUnit,
		PropertyTagThumbnailTransferFunction,
		PropertyTagThumbnailSoftwareUsed,
		PropertyTagThumbnailDateTime,
		PropertyTagThumbnailArtist,
		PropertyTagThumbnailWhitePoint,
		PropertyTagThumbnailPrimaryChromaticities,
		PropertyTagThumbnailYCbCrCoefficients,
		PropertyTagThumbnailYCbCrSubsampling,
		PropertyTagThumbnailYCbCrPositioning,
		PropertyTagThumbnailRefBlackWhite,
		PropertyTagThumbnailCopyRight,
		PropertyTagLuminanceTable,
		PropertyTagChrominanceTable,
		PropertyTagFrameDelay,
		PropertyTagLoopCount,
		PropertyTagPixelUnit,
		PropertyTagPixelPerUnitX,
		PropertyTagPixelPerUnitY,
		PropertyTagPaletteHistogram,
		PropertyTagExifExposureTime,
		PropertyTagExifFNumber,
		PropertyTagExifExposureProg,
		PropertyTagExifSpectralSense,
		PropertyTagExifISOSpeed,
		PropertyTagExifOECF,
		PropertyTagExifVer,
		PropertyTagExifDTOrig,
		PropertyTagExifDTDigitized,
		PropertyTagExifCompConfig,
		PropertyTagExifCompBPP,
		PropertyTagExifShutterSpeed,
		PropertyTagExifAperture,
		PropertyTagExifBrightness,
		PropertyTagExifExposureBias,
		PropertyTagExifMaxAperture,
		PropertyTagExifSubjectDist,
		PropertyTagExifMeteringMode,
		PropertyTagExifLightSource,
		PropertyTagExifFlash,
		PropertyTagExifFocalLength,
		PropertyTagExifMakerNote,
		PropertyTagExifUserComment,
		PropertyTagExifDTSubsec,
		PropertyTagExifDTOrigSS,
		PropertyTagExifDTDigSS,
		PropertyTagExifFPXVer,
		PropertyTagExifColorSpace,
		PropertyTagExifPixXDim,
		PropertyTagExifPixYDim,
		PropertyTagExifRelatedWav,
		PropertyTagExifInterop,
		PropertyTagExifFlashEnergy,
		PropertyTagExifSpatialFR,
		PropertyTagExifFocalXRes,
		PropertyTagExifFocalYRes,
		PropertyTagExifFocalResUnit,
		PropertyTagExifSubjectLoc,
		PropertyTagExifExposureIndex,
		PropertyTagExifSensingMethod,
		PropertyTagExifFileSource,
		PropertyTagExifSceneType,
		PropertyTagExifCfaPattern,
		PropertyTagGpsVer,
		PropertyTagGpsLatitudeRef,
		PropertyTagGpsLatitude,
		PropertyTagGpsLongitudeRef,
		PropertyTagGpsLongitude,
		PropertyTagGpsAltitudeRef,
		PropertyTagGpsAltitude,
		PropertyTagGpsGpsTime,
		PropertyTagGpsGpsSatellites,
		PropertyTagGpsGpsStatus,
		PropertyTagGpsGpsMeasureMode,
		PropertyTagGpsGpsDop,
		PropertyTagGpsSpeedRef,
		PropertyTagGpsSpeed,
		PropertyTagGpsTrackRef,
		PropertyTagGpsTrack,
		PropertyTagGpsImgDirRef,
		PropertyTagGpsImgDir,
		PropertyTagGpsMapDatum,
		PropertyTagGpsDestLatRef,
		PropertyTagGpsDestLat,
		PropertyTagGpsDestLongRef,
		PropertyTagGpsDestLong,
		PropertyTagGpsDestBearRef,
		PropertyTagGpsDestBear,
		PropertyTagGpsDestDistRef,
		PropertyTagGpsDestDist,
		QPropertyTagSubjectArea,
		QPropertyTagSubSecTime,
		QPropertyTagSubSecTimeOriginal,
		QPropertyTagCustomRendered,
		QPropertyTagExposureMode,
		QPropertyTagWhiteBalance,
		QPropertyTagDigitalZoomRatio,
		QPropertyTagFocalLengthIn35mmFilm,
		QPropertyTagSceneCaptureType,
		QPropertyTagGainControl,
		QPropertyTagContrast,
		QPropertyTagSaturation,
		QPropertyTagSharpness,
		QPropertyTagDeviceSettingDescription,
		QPropertyTagSubjectDistanceRange,
		QPropertyTagImageUniqueID
	};
	const UINT nTags = sizeof(idPropTags)/sizeof(UINT);

	CString s;

	for (UINT i = 0; i < nTags; i++)
		if (tag == idPropTags[i])
		{
			VERIFY(s.LoadString(QDS_PROPTAGS + i));
			break;
		}
	if (i >= nTags) s.Format(QDS_UNKNOWNPROP, tag);
	return s;
}

/* static */
CString QImageProperties::FileString(LPCTSTR pFilePath, LPCTSTR colon)
{
	LPCTSTR endl = _T("\r\n");

	CString sFile;
	if (pFilePath != NULL && _tcslen(pFilePath) > 0)
	{
		VERIFY(sFile.LoadString(QDS_FILE));
		// If you get an ASSERT here, chances are that you didn't include the
		// QImageProperties resources. These are in QImagePoperties.rc.
		// This file should be included in the application's main .rc-file.

		sFile += colon;
		sFile += pFilePath;
		sFile += endl;

		WIN32_FILE_ATTRIBUTE_DATA fileData;
		if (::GetFileAttributesEx(pFilePath, GetFileExInfoStandard, & fileData))
		{
			::CStringAppendFormat(sFile, QDS_CREATED);
			sFile += colon;
			sFile += FileTimeToString(fileData.ftCreationTime);
			sFile += endl;

			::CStringAppendFormat(sFile, QDS_ACCESSED);
			sFile += colon;
			sFile += FileTimeToString(fileData.ftLastAccessTime);
			sFile += endl;

			::CStringAppendFormat(sFile, QDS_LASTWRITTEN);
			sFile += colon;
			sFile += FileTimeToString(fileData.ftLastWriteTime);
			sFile += endl;

			ULARGE_INTEGER u;
			u.LowPart = fileData.nFileSizeLow;
			u.HighPart = fileData.nFileSizeHigh;
			::CStringAppendFormat(sFile, QDS_FILESIZE, u.QuadPart);
			sFile += endl;

		}
	}
	return sFile;
}

/* static */
CString QImageProperties::RealToString(double r)
{
	CString s;
	s.Format(_T("%f"), r);
	s.TrimRight(_T('0'));
	if (s[s.GetLength() - 1] == _T('.')) s += _T('0');
	return s;
}

/* static */
CString QImageProperties::FileTimeToString(FILETIME& t)
{
	FILETIME tLocal;
	SYSTEMTIME st;
	::FileTimeToLocalFileTime(& t, & tLocal);
	::FileTimeToSystemTime(& tLocal, & st);

	CString s;

	// same format as Exif date fields
	s.Format(_T("%04d:%02d:%02d %02d:%02d:%02d"), st.wYear, st.wMonth, st.wDay,
			st.wHour, st.wMinute, st.wSecond);
	return s;
}

//===========================
// QImagePropertiesDlg dialog

IMPLEMENT_DYNAMIC(QImagePropertiesDlg, CDialog)
QImagePropertiesDlg::QImagePropertiesDlg(CWnd* pParent /*=NULL*/)
: CDialog(QDD_IMAGEPROPS, pParent)
, m_Description(_T(""))
, m_Author(_T(""))
, m_Copyright(_T(""))
, m_Props(_T(""))
{
}

QImagePropertiesDlg::~QImagePropertiesDlg()
{
}

void QImagePropertiesDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Text(pDX, QDC_DESCRIPTION, m_Description);
	DDX_Text(pDX, QDC_AUTHOR, m_Author);
	DDX_Text(pDX, QDC_COPYRIGHT, m_Copyright);
	DDX_Text(pDX, QDC_PROPS, m_Props);
}

bool QImagePropertiesDlg::Show(Image * pSource, Image * pDest, LPCTSTR pFilePath)
{
	if (! pSource) return false;
	if (! pDest) pDest = pSource;

	QImageProperties propsSource(pSource);

	propsSource.GetProperty(PropertyTagArtist, m_Author);
	propsSource.GetProperty(PropertyTagImageDescription, m_Description);
	propsSource.GetProperty(PropertyTagCopyright, m_Copyright);

	if (pFilePath) m_Props = QImageProperties::FileString(pFilePath);

	// Zero-terminated list of property tags to be skipped by AllProperties().
	// We skip them, because they are already in the other controls.
	PROPID skipList[] =
	{
		PropertyTagArtist,
		PropertyTagImageDescription,
		PropertyTagCopyright,
		0
	};
	m_Props += propsSource.AllProperties(skipList);

	bool r = (DoModal() == IDOK);

	if (r)
	{
		QImageProperties propsDest(pDest);

		propsDest.SetProperty(PropertyTagArtist, m_Author);
		propsDest.SetProperty(PropertyTagImageDescription, m_Description);
		propsDest.SetProperty(PropertyTagCopyright, m_Copyright);
	}
	return r;
}

BEGIN_MESSAGE_MAP(QImagePropertiesDlg, CDialog)
END_MESSAGE_MAP()
