#pragma once

// QImageProperties and QImagePropertiesDlg
//
// Classes to work with GDI+ Image properties.
// Use at your own risk. Comments welcome.
//
// Version 1.0 (c) 2003, 10/20/2003, Sjaak Priester, Amsterdam.
// mailto:sjaak@sjaakpriester.nl
//
// These classes need the resources in QImagePoperties.rc.
// This file should be included in the main applications .rc-file.
// VC++ 6.0: choose View => Resource Includes.
// VC++ 7.0: in Resource View, right-click on *.rc, choose Resource Includes.
// Add #include "QImagePoperties.rc" under Compile-time Directives.
//
// Because VC++ 6.0 does not support function templates,
// some of the functions only compile under VC++ 7.0 and later.
// These are in the if-else blocks with condition _MSC_VER >= 1300.

class QImageProperties
{
public:
	QImageProperties(Image * pImage);
	// Constructor; pImage is the associated Image.
	virtual ~QImageProperties(void);

	void Copy(const QImageProperties& source, PROPID * pSkipList = NULL,
		bool bSkipThumbnail = true);
	// Copy the stored properties from source.
	//
	// Parameters:
	// - pSkipList		if non-NULL, points to a zero-terminated list of property
	//					tags which are skipped in the copying process;
	// - bSkipThumbnail	if true, the range of Thumbnail properties are skipped.

	BYTE * GetPropertyBytes(PROPID tag) const;

#if _MSC_VER >= 1300
	bool GetProperty(PROPID tag, BYTE& d) const
		{ return GetSingleProperty<BYTE, PropertyTagTypeByte>(tag, d); }
	bool GetProperty(PROPID tag, short& d) const
		{ return GetSingleProperty<short, PropertyTagTypeShort>(tag, d); }
	bool GetProperty(PROPID tag, UINT& d) const
		{ return GetSingleProperty<UINT, PropertyTagTypeLong>(tag, d); }
	bool GetProperty(PROPID tag, int& d) const
		{ return GetSingleProperty<int, PropertyTagTypeSLONG>(tag, d); }
#endif

	bool GetProperty(PROPID tag, REAL& d) const;
	bool GetProperty(PROPID tag, CString& s) const;
	// Methods to retrieve a property. They only work with single properties.
	// Return: true if succeeded, false if failed (meaning: the property is not
	// supported, the Image doesn't have this property, or tag denotes a multiple property).

#if _MSC_VER >= 1300
	bool SetProperty(PROPID tag, BYTE d)
		{ return SetSingleProperty<BYTE, PropertyTagTypeByte>(tag, d); }
	bool SetProperty(PROPID tag, short d)
		{ return SetSingleProperty<short, PropertyTagTypeShort>(tag, d); }
	bool SetProperty(PROPID tag, UINT d)
		{ return SetSingleProperty<UINT, PropertyTagTypeLong>(tag, d); }
	bool SetProperty(PROPID tag, int d)
		{ return SetSingleProperty<int, PropertyTagTypeSLONG>(tag, d); }
	bool SetProperty(PROPID tag, UINT num, UINT denom)
		{ return SetSingleRatio<UINT, PropertyTagTypeRational>(tag, num, denom); }
	bool SetProperty(PROPID tag, int num, int denom)
		{ return SetSingleRatio<int, PropertyTagTypeSRational>(tag, num, denom); }
	bool SetProperty(PROPID tag, REAL d);
#endif

	bool SetProperty(PROPID tag, CString s);
	// Methods to retrieve a property. Return: like Get...() counterparts.


	CString PropertyString(PROPID tag, LPCTSTR colon = _T(": ")) const;
	// Get a string of the property name + colon + a textual representation of the
	// property value. If the property is not available (not supported or not set),
	// just the name + colon is returned.

	CString AllProperties(PROPID * pSkipList = NULL, LPCTSTR colon = _T(": ")) const;
	// Get a string of all the (inherent and stored) properties.
	// If pSkipList is non-NULL, it points to a zero-terminated list of property
	// tags which are excluded from the resulting string.

	static CString FileString(LPCTSTR pFilePath, LPCTSTR colon = _T(": "));
	// Return string with file info (name, length, dates).
	// if pFilePath == NULL or pointer to empty string, return empty string.

	static CString PropertyName(PROPID tag);
	// Get a string with the name of the property with id tag.

protected:
	Image *m_pImage;

#if _MSC_VER >= 1300
	// Only succeeds for single value of T
	template <class T, WORD Y>
	bool GetSingleProperty(PROPID tag, T& d) const
	{
		BYTE * pBytes = GetPropertyBytes(tag);

		bool r = false;
		if (pBytes)
		{
			PropertyItem * pProp = (PropertyItem *) pBytes;
			if (pProp->type == Y && pProp->length == sizeof(T))
			{
				T * p = (T *) pProp->value;
				d = * p;
				r = true;
			}
		}
		delete[] pBytes;
		return r;
	};

	template <class T, WORD Y>
	bool SetSingleProperty(PROPID tag, T d)
	{
		PropertyItem prop;
		prop.type = Y;
		prop.id = tag;
		prop.length = sizeof(T);
		prop.value = & d;
		return (m_pImage->SetPropertyItem(& prop) == Ok);
	};

	template <class T, WORD Y>
	bool SetSingleRatio(PROPID tag, T num, T denom)
	{
		T d[2];
		d[0] = num;
		d[1] = denom;

		PropertyItem prop;
		prop.type = Y;
		prop.id = tag;
		prop.length = 2 * sizeof(T);
		prop.value = d;
		return (m_pImage->SetPropertyItem(& prop) == Ok);
	};
#endif

	static CString GuidString(GUID& guid);
	static CString FileTimeToString(FILETIME& t);
	static CString RealToString(double r);
};

class QImagePropertiesDlg : public CDialog
{
public:
	QImagePropertiesDlg(CWnd* pParent = NULL);
	virtual ~QImagePropertiesDlg();

	bool Show(Image * pSource, Image * pDest = NULL, LPCTSTR pFilePath = NULL);
	// Show dialog with properties of pSource.
	// If pDest is not NULL, updated properties are stored in pDest. If NULL,
	// they are stored in pSource. pDest and pSource may point to the same Image.
	// If pFilePath is not NULL, file information is included in the dialog.

	CString m_Description;
	CString m_Author;
	CString m_Copyright;
	CString m_Props;

protected:
	virtual void DoDataExchange(CDataExchange* pDX);

	DECLARE_DYNAMIC(QImagePropertiesDlg)
	DECLARE_MESSAGE_MAP()
};
