// TinterView.cpp
//
// Part of Tinter, a small app to demonstrate QColorMatrix, QGaussMatrix, QImageProperties.
// Use at your own risk. Comments welcome.
//
// (c) 2003-2004, Sjaak Priester, Amsterdam.
// mailto:sjaak@sjaakpriester.nl
// www.sjaakpriester.nl
//
// Version 1.3:
// Gif preview added
// Size handling added
//
// Version 1.2:
// Zoom capabilities added
// Twain support added
// Negative control added
// Rotate and flip handling enhanced
// Numerous small changes
//
// Version 1.1:
// View toggle original/modified added
// Toggle border added
// QGaussFilter support added for Unsharp Mask and Blur
// View picture properties command added
// Rotate and Flip commands added
// One save option (JPEG quality) added
// Properties copied at Save
// Image encoder functionality wrapped in class QImageEncoder
// Indicators added for size and scale
// Tint controls added
//
// Known problems:
// Twain: gray level and line art scanning don't work correctly. Bitmap data is shifted.
//		Bug in Bitmap(BITMAPINFO *) ctor?
//
// Bugs fixed:
// 09/16/2003: no more SetDirty() etc. if no picture
// 09/11/2003: Invalidate() added to OnUpdate (small gif file didn't erase bigger tiff)
//
// Reference & thanks:
// Rajiv Ramachandran (twain support) http://www.codeguru.com/multimedia/twaintest.shtml

#include "stdafx.h"
#include "Resource.h"

#include "TinterDoc.h"
#include "TinterView.h"
#include "SliderBar.h"
#include "GifPreviewBar.h"
#include "SizeDlg.h"

#include "QImageEncoder.h"
#include "QImageProperties.h"
#include "QColorQuantizer.h"
#include "Int.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#define ID_ZOOM_FIRST		ID_ZOOM_12
#define ID_ZOOM_LAST		ID_ZOOM_800

#define ZOOM_100			(ID_ZOOM_100 - ID_ZOOM_FIRST)

// Helper function: makes toFit fit inside room.
// If toFit is smaller than room, return toFit unchanged.
// Otherwise, scale down and return new, smaller size.
CSize FitSize(CSize toFit, CSize room)
{
	if (toFit.cx > room.cx || toFit.cy > room.cy)		// Picture is greater than window, scale.
								// Smaller pictures retain their size.
	{
		if (toFit.cx * room.cy > room.cx * toFit.cy)	// width determines scale
		{
			toFit.cy = MulDiv(toFit.cy, room.cx, toFit.cx);
			toFit.cx = room.cx;
		}
		else					// height determines scale
		{
			toFit.cx = MulDiv(toFit.cx, room.cy, toFit.cy);
			toFit.cy = room.cy;
		}
	}
	return toFit;
}

// CTinterView

IMPLEMENT_DYNCREATE(CTinterView, QZoomView)

BEGIN_MESSAGE_MAP(CTinterView, QZoomView)
	ON_COMMAND(ID_FILE_PRINT, QZoomView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, QZoomView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, QZoomView::OnFilePrintPreview)
	ON_MESSAGE(QM_SLIDERBAR, OnSliderBar)
	ON_MESSAGE(QM_GAUSSFILTER, OnGaussFilter)
	ON_COMMAND(ID_EDIT_UNDO, OnSetNeutral)
	ON_UPDATE_COMMAND_UI(ID_FILE_PRINT, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_FILE_PRINT_PREVIEW, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE_AS, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_EDIT_UNDO, OnUpdateGeneral)
	ON_COMMAND(ID_FILE_SAVEOPTIONS, OnFileSaveoptions)
	ON_COMMAND(ID_VIEW_MODIFIED, OnViewModified)
	ON_UPDATE_COMMAND_UI(ID_VIEW_MODIFIED, OnUpdateViewModified)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_CALCULATING, OnUpdateCalculating)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_MODIFIED, OnUpdateModified)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_SCALE, OnUpdateScale)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_SIZE, OnUpdateSize)
	ON_COMMAND(ID_VIEW_BORDER, OnViewBorder)
	ON_UPDATE_COMMAND_UI(ID_VIEW_BORDER, OnUpdateViewBorder)
	ON_COMMAND(ID_FILE_USEIMAGERESOLUTION, OnFileUseimageresolution)
	ON_UPDATE_COMMAND_UI(ID_FILE_USEIMAGERESOLUTION, OnUpdateFileUseimageresolution)
	ON_UPDATE_COMMAND_UI_RANGE(ID_ROTATE_90, ID_ROTATE_270, OnUpdateRotateFlip)
	ON_UPDATE_COMMAND_UI_RANGE(ID_FLIP_HORIZONTAL, ID_FLIP_VERTICAL, OnUpdateRotateFlip)
	ON_COMMAND(ID_VIEW_ZOOMIN, OnViewZoomin)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOMIN, OnUpdateViewZoomin)
	ON_COMMAND(ID_VIEW_ZOOMOUT, OnViewZoomout)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOMOUT, OnUpdateViewZoomout)
	ON_COMMAND(ID_VIEW_ZOOMTOWINDOW, OnViewZoomtowindow)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOMTOWINDOW, OnUpdateGeneral)
	ON_COMMAND(ID_VIEW_ACTUALSIZE, OnViewActualsize)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ACTUALSIZE, OnUpdateGeneral)
	ON_COMMAND(ID_TOOL_PLUSLOUPE, OnToolPlusloupe)
	ON_UPDATE_COMMAND_UI(ID_TOOL_PLUSLOUPE, OnUpdateToolPlusloupe)
	ON_COMMAND(ID_TOOL_MINUSLOUPE, OnToolMinusloupe)
	ON_UPDATE_COMMAND_UI(ID_TOOL_MINUSLOUPE, OnUpdateToolMinusloupe)
	ON_COMMAND(ID_TOOL_HAND, OnToolHand)
	ON_UPDATE_COMMAND_UI(ID_TOOL_HAND, OnUpdateToolHand)
	ON_COMMAND_RANGE(ID_ZOOM_FIRST, ID_ZOOM_LAST, OnZoom)
	ON_UPDATE_COMMAND_UI_RANGE(ID_ZOOM_FIRST, ID_ZOOM_LAST, OnUpdateZoom)
	ON_COMMAND(ID_ZOOM_OTHER, OnZoomOther)
	ON_UPDATE_COMMAND_UI(ID_ZOOM_OTHER, OnUpdateGeneral)
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_VIEW_PROPERTIES, OnViewProperties)
	ON_UPDATE_COMMAND_UI(ID_VIEW_PROPERTIES, OnUpdateGeneral)
	ON_COMMAND(ID_VIEW_BACKGROUNDCOLOR, OnViewBackgroundcolor)
	ON_COMMAND(ID_VIEW_BORDERCOLOR, OnViewBordercolor)
	ON_COMMAND(ID_FILE_SELECTSOURCE, OnFileSelectsource)
	ON_UPDATE_COMMAND_UI(ID_FILE_SELECTSOURCE, OnUpdateFileSelectsource)
	ON_WM_CREATE()
	ON_COMMAND(ID_FLIP_HORIZONTAL, OnFlipHorizontal)
	ON_COMMAND(ID_FLIP_VERTICAL, OnFlipVertical)
	ON_COMMAND(ID_ROTATE_180, OnRotate180)
	ON_COMMAND(ID_ROTATE_270, OnRotate270)
	ON_COMMAND(ID_ROTATE_90, OnRotate90)
	ON_COMMAND(ID_VIEW_GIFPREVIEW, OnViewGifpreview)
	ON_UPDATE_COMMAND_UI(ID_VIEW_GIFPREVIEW, OnUpdateViewGifpreview)
	ON_UPDATE_COMMAND_UI(ID_INDICATOR_GIF, OnUpdateGifpreview)
	ON_COMMAND(IDC_CLOSEPREVIEW, OnCloseGifPreview)
	ON_COMMAND_RANGE(IDC_WEBSAFE, IDC_OCTREE, OnGifMode)
	ON_UPDATE_COMMAND_UI_RANGE(IDC_WEBSAFE, IDC_OCTREE, OnUpdateGifMode)
	ON_COMMAND(ID_EDIT_SIZE, OnEditSize)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SIZE, OnUpdateGeneral)
END_MESSAGE_MAP()

// CTinterView construction/destruction

CTinterView::CTinterView()
: m_pModified(NULL)
, m_pOriginal(NULL)
, m_pGif(NULL)
, m_SaveQuality(75)		// seems to be the default value
, m_bViewModified(true)
, m_Gamma(1.0f)
, m_ColorMatrix(QGaussFilter::MatYUVtoRGB)
, m_bBorder(false)
, m_bImageResolution(false)
, m_szDoc(100, 100)
, m_colBackGround(::GetSysColor(COLOR_WINDOW))
, m_colBorder(::GetSysColor(COLOR_WINDOWTEXT))
, m_bGifOctree(true)
, m_bGifPreview(false)
{
}

CTinterView::~CTinterView()
{
	delete m_pModified;
	delete m_pOriginal;
	delete m_pGif;
}

BOOL CTinterView::PreCreateWindow(CREATESTRUCT& cs)
{
	return QZoomView::PreCreateWindow(cs);
}

// CTinterView drawing

// We don't draw the image directly to the screen, but from a buffer bitmap.
// We do this for two reasons:
// - GDI+ keeps any image file locked as long as it is on the screen;
// - we keep the buffer bitmap in the YUV color space in stead of RGB,
//		thus making filter operations faster.
void CTinterView::OnDraw(CDC* pDC)
{
	if (m_pModified)
	{
		Graphics g(pDC->GetSafeHdc());
		g.SetInterpolationMode(GetZoom() > 1.0f ? InterpolationModeNearestNeighbor 
			: InterpolationModeHighQualityBicubic);

		CRect rcClip;
		pDC->GetClipBox(rcClip);
		rcClip.InflateRect(1, 1);

		Rect rcgClip(rcClip.left, rcClip.top, rcClip.Width(), rcClip.Height());
		Rect rcDraw;

		rcDraw.Intersect(rcDraw, m_Rect, rcgClip);
		int x = rcDraw.X - m_Rect.X;
		int y = rcDraw.Y - m_Rect.Y;

		if (m_bViewModified)
		{
			// Draw the buffer image to the screen, using
			// m_ColorMatrix
			ImageAttributes attr;
			attr.SetColorMatrix(& m_ColorMatrix);
			attr.SetGamma(m_Gamma);

			g.DrawImage(m_pModified,
				rcDraw,
				x, y,
				rcDraw.Width, rcDraw.Height,
				UnitPixel,
				& attr);
		}
		else
		{
			if (m_bGifPreview)
			{
				ASSERT(m_pGif);

				// Draw indexed bitmap to the screen

				g.DrawImage(m_pGif,
					rcDraw,
					x, y,
					rcDraw.Width, rcDraw.Height,
					UnitPixel);
			}
			else
			{
				ASSERT(m_pOriginal);

				// Draw the orignal image to the screen, *not* using
				// m_ColorMatrix

				ImageAttributes attr;
				attr.SetColorMatrix(& QGaussFilter::MatYUVtoRGB);

				g.DrawImage(m_pOriginal,
					rcDraw,
					x, y,
					rcDraw.Width, rcDraw.Height,
					UnitPixel,
					& attr);
			}
		}

		if (m_bBorder)
		{
			Color col;
			col.SetFromCOLORREF(m_colBorder);
			Pen pen(col);
			g.DrawRectangle(& pen, m_Rect);
		}
	}
	else	// no image loaded
	{
		CGdiObject * pOldFont = pDC->SelectStockObject(DEFAULT_GUI_FONT);

		CString s;
		s.LoadString(IDS_NOPICTURE);
		CRect rcClient;
		GetClientRect(rcClient);
		CPoint center = rcClient.CenterPoint();

		// Text color is black for light background colors, white for dark ones
		int bgLum = 31 * GetRValue(m_colBackGround)
			+ 61 * GetGValue(m_colBackGround)
			+ 8 * GetBValue(m_colBackGround);

		COLORREF colPen = bgLum > (60 * 255) ? RGB(0, 0, 0) : RGB(255, 255, 255);

		COLORREF colOld = pDC->SetTextColor(colPen);
		pDC->SetBkMode(TRANSPARENT);
		pDC->SetTextAlign(TA_CENTER | TA_BASELINE);
		pDC->TextOut(center.x, center.y, s);
		pDC->SetTextColor(colOld);

		if (pOldFont) pDC->SelectObject(pOldFont);
	}
}

BOOL CTinterView::OnEraseBkgnd(CDC* pDC)
{
	CRect rcClip;
	pDC->GetClipBox(rcClip);

	if (m_pModified)
	{
		CClientDC dc(this);

		OnPrepareDC(& dc);

		CRgn rgn;
		rgn.CreateRectRgnIndirect(rcClip);

		CRect rcDraw(m_Rect.GetLeft(), m_Rect.GetTop(), m_Rect.GetRight(), m_Rect.GetBottom());
		dc.LPtoDP(rcDraw);
		rcDraw.DeflateRect(1, 1);

		CRgn rgnDraw;
		rgnDraw.CreateRectRgnIndirect(rcDraw);

		if (rgn.CombineRgn(& rgn, & rgnDraw, RGN_DIFF) != NULLREGION)
		{
			CBrush br(m_colBackGround);
			pDC->FillRgn(& rgn, & br);
		}
	}
	else pDC->FillSolidRect(rcClip, m_colBackGround);
	return TRUE;
}

//===============================
// printing support

// Print Preview and GDI+ don't go together very well, I assume because
// MFC knows nothing about GDI+. I simply left out the menu item.
void CTinterView::OnPrint(CDC* pDC, CPrintInfo* pInfo)
{
	if (m_pModified)
	{
		// Scale to prepare for UnitDocument (1/300 inch), assuming the picture
		// resolution equals the screen resolution.
		// In theory it is preferable to use the resolution of the picture itself.
		// However, some pictures store a weird resolution and end up way to small,
		// so I used the screen resolution as a default.
		// Check the 'Use image resolution' in the File Menu to use the picture resolution.

		UINT w, h;

		if (m_bImageResolution)
		{
			w = ::MulDiv(m_Rect.Width, 300, (int) m_pModified->GetHorizontalResolution());
			h = ::MulDiv(m_Rect.Height, 300, (int) m_pModified->GetVerticalResolution());
		}
		else
		{
			CWindowDC dcScreen(NULL);

			w = ::MulDiv(m_Rect.Width, 300, dcScreen.GetDeviceCaps(LOGPIXELSX));
			h = ::MulDiv(m_Rect.Height,	300, dcScreen.GetDeviceCaps(LOGPIXELSY));
		}

		CRect rcDraw(pInfo->m_rectDraw);
		rcDraw.DeflateRect(10, 10);		// small margin
		CSize szPic = FitSize(CSize(w, h), rcDraw.Size());

		// Center on the page
		CRect rcPicture(CPoint(0, 0), szPic);
		CSize szOffset = pInfo->m_rectDraw.CenterPoint() - rcPicture.CenterPoint();
		
		Rect rc(szOffset.cx, szOffset.cy, szPic.cx, szPic.cy);

		ImageAttributes attr;
		attr.SetColorMatrix(& m_ColorMatrix);
		attr.SetGamma(m_Gamma);

		Graphics g(pDC->GetSafeHdc());
		g.SetPageUnit(UnitDocument);

		g.DrawImage(m_pModified,
			rc,
			0, 0,
			m_Rect.Width, m_Rect.Height,
			UnitPixel,
			& attr);

		if (m_bBorder)
		{
			Color col;
			col.SetFromCOLORREF(m_colBorder);
			Pen pen(col);
			g.DrawRectangle(& pen, rc);
		}
	}
}

BOOL CTinterView::OnPreparePrinting(CPrintInfo* pInfo)
{
	return DoPreparePrinting(pInfo);
}

void CTinterView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
}

void CTinterView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
}

// overrides

void CTinterView::OnPrepareDC(CDC* pDC, CPrintInfo* pInfo)
{
	// Don't call base class if no picture is loaded.
	if (m_pModified) QZoomView::OnPrepareDC(pDC, pInfo);
}

void CTinterView::OnInitialUpdate()
{
	QZoomView::OnInitialUpdate();
	SetScrollSizes(MM_TEXT, m_szDoc);
	if (m_pModified) FitInWindow();
}

// In a slight variation of MFC's normal doc-view architecture, the document uses the update
// mechanism only to instruct the view to load or save a file.
// Normally these operations are the responsibility of the document, but because we need GDI+,
// we let the view handle them.
void CTinterView::OnUpdate(CView* /*pSender*/, LPARAM lHint, CObject* pHint)
{
	switch (lHint)
	{
	case HINT_OPEN:		// defined in CTinterDoc.h
		{
			// Opening a file is easy!
			CString sPath((LPCTSTR) pHint);
			Bitmap bm(sPath.AllocSysString());
			Status stat = bm.GetLastStatus();
			if (stat == Ok)
			{
				SetBitmap(& bm);
				QImageProperties propSource(& bm);
				QImageProperties propDest(m_pOriginal);
				propDest.Copy(propSource, false);
			}
			else
			{
				CString s;
				s.Format(IDS_LOADERROR, stat, ::GetLastError());
				::AfxMessageBox(s);
			}
		}
		break;
	case HINT_SAVE:		// Saving a file is less easy!
		{
			Status stat = Save((LPCTSTR) pHint);
			if (stat != Ok)
			{
				CString s;
				s.Format(IDS_SAVEERROR, stat, ::GetLastError());
				::AfxMessageBox(s);
			}
		}
		break;
	case HINT_NEW:
		Acquire();
		break;
	default:
		Invalidate();
		break;
	}
}

//=======================================
// slider bar

// Called when CSliderBar notifies us about a value change.
LRESULT CTinterView::OnSliderBar(WPARAM wParam, LPARAM /*lParam*/)
{
	if (! m_pModified) return 0;

	CSliderBar * pSliderBar = (CSliderBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_SLIDERBAR);
	ASSERT(pSliderBar);

	switch (wParam)
	{
	case IDC_FILTER:
	case IDC_RADIUS:
	case IDC_DEPTH:
	case IDC_LUMINANCE:
		{
			UINT flags = pSliderBar->GetLuminance()
				? QGaussFilter::PlaneGreen : QGaussFilter::PlaneAll;
			switch (pSliderBar->GetFilterMode())
			{
			case CSliderBar::FilterOff:
				FilterOff();
				Invalidate(FALSE);
				break;
			case CSliderBar::FilterUnsharpMask:
				m_Filter.MakeUnsharpMask(m_pOriginal,
					pSliderBar->GetRadius(),
					pSliderBar->GetDepth(),
					this,
					QM_GAUSSFILTER,
					NULL,
					flags);
				break;
			case CSliderBar::FilterBlur:
				m_Filter.MakeBlur(m_pOriginal,
					pSliderBar->GetRadius(),
					this,
					QM_GAUSSFILTER,
					NULL,
					flags);
				break;
			default:
				ASSERT(FALSE);
				break;
			}
		}

		// If we handled a change in a filter parameter we're ready now.
		return 0;

	case 0:		// Neutral button clicked
		FilterOff();
		break;

	default:
		break;
	}

	// Image is kept in YUV color space; compensate
	m_ColorMatrix = QGaussFilter::MatYUVtoRGB;

	if (pSliderBar->GetNegative())
	{
		m_ColorMatrix.ScaleColors(-1.0f, MatrixOrderAppend);
		m_ColorMatrix.TranslateColors(1.0f, MatrixOrderAppend);
	}

	// Update color matrix with current values
	m_ColorMatrix.ScaleColors(pSliderBar->GetContrast(), MatrixOrderAppend);
	m_ColorMatrix.TranslateColors(pSliderBar->GetBrightness(), MatrixOrderAppend);
	m_ColorMatrix.SetSaturation(pSliderBar->GetSaturation(), MatrixOrderAppend);
	m_ColorMatrix.RotateHue(pSliderBar->GetHue());
	m_ColorMatrix.SetTint(pSliderBar->GetTintHue(), pSliderBar->GetTintAmount());

	// Okay, gamma has nothing to do with color matrix operations,
	// but we throw it in for free.
	m_Gamma = pSliderBar->GetGamma();

    m_bViewModified = true;

	// Invalidate, not erasing background (important for smooth update)
	Invalidate(FALSE);

	InvalidateDocument();
	return 0;
}

// Called when QGaussFilter is ready with calculating the filtered bitmap
// in a separate thread.
LRESULT CTinterView::OnGaussFilter(WPARAM /*wParam*/, LPARAM lParam)
{
	// wParam has the resulting bitmap
	Bitmap * pResult = (Bitmap *) lParam;
	if (pResult)
	{
		delete m_pModified;
		m_pModified = pResult;
	    m_bViewModified = true;
		Invalidate(FALSE);
		InvalidateDocument();
	}
	return 0;
}

void CTinterView::FitInWindow(void)
{
	CRect rcClient;
	GetClientRect(rcClient);
	int margin = ::GetSystemMetrics(SM_CXFRAME);
	rcClient.DeflateRect(margin, margin);

	if (m_szDoc.cx > rcClient.Width() || m_szDoc.cy > rcClient.Height()) ZoomToWindow();
	else ZoomTo(1.0f);
}

void CTinterView::FilterOff(void)
{
	if (m_pModified)
	{
		ASSERT(m_pOriginal);
		delete m_pModified;
		m_pModified = m_pOriginal->Clone(0, 0, m_Rect.Width, m_Rect.Height, PixelFormat24bppRGB);
	}
}

Status CTinterView::Save(LPCTSTR pPath)
{
	if (! m_pModified) return GenericError;
	ASSERT(m_pOriginal);

	QImageProperties propSource(m_pOriginal);
	QImageProperties propDest(m_pModified);
	propDest.Copy(propSource);

	CString s;
	s.LoadString(IDS_APPNAME);
	propDest.SetProperty(PropertyTagSoftwareUsed, s);

	QImageEncoder encoder(pPath);
	encoder.m_Quality = m_SaveQuality;

	Status stat(GenericError);
	Rect rc(0, 0, m_Rect.Width, m_Rect.Height);

	if (encoder.GetMimeType() == _T("image/gif"))
	{
		if (m_pGif == NULL) CreateGif();
		if (m_pGif) stat = encoder.Save(m_pGif, pPath);
	}
	else
	{
		// Apply m_ColorMatrix and gamma, and convert from YUV to RGB
		ImageAttributes attr;
		attr.SetColorMatrix(& m_ColorMatrix);
		attr.SetGamma(m_Gamma);

		// Although not documented as such, printing a bitmap on itself works.
		Graphics g(m_pModified);
		g.DrawImage(m_pModified,
			rc,
			0, 0, rc.Width, rc.Height,
			UnitPixel,
			& attr);

		// Finally, save it!
		stat = encoder.Save(m_pModified, pPath);

		// Convert back to YUV
		QGaussFilter::RGBtoYUV(m_pModified);
	}

	if (stat == Ok)
	{
		// New original = modified
		delete m_pOriginal;
		m_pOriginal = m_pModified->Clone(rc, m_pModified->GetPixelFormat());

		OnSetNeutral();
		GetDocument()->SetModifiedFlag(FALSE);
	}

	return stat;
}

//========================
// command handlers

// general update handler; enables if picture loaded
void CTinterView::OnUpdateGeneral(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pModified != NULL);
}

// File menu

void CTinterView::OnFileSaveoptions()
{
	COptionsDlg dlg(this);
	dlg.m_JPEGQuality = m_SaveQuality;
	dlg.m_GifMode = m_bGifOctree;
	if (dlg.DoModal() == IDOK)
	{
		m_SaveQuality = dlg.m_JPEGQuality;

		bool bGifModeChanged = (m_bGifOctree != (dlg.m_GifMode == TRUE));
		m_bGifOctree = (dlg.m_GifMode == TRUE);

		if (bGifModeChanged && m_bGifPreview)
		{
			CreateGif();
			Invalidate(FALSE);
		}
	}
}

void CTinterView::OnFileUseimageresolution()
{
	m_bImageResolution ^= true;
}

void CTinterView::OnUpdateFileUseimageresolution(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_bImageResolution);
}

// Edit menu

// also called at File New (= Acquire) or File Open 
void CTinterView::OnSetNeutral()
{
	CSliderBar * pSliderBar = (CSliderBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_SLIDERBAR);
	ASSERT(pSliderBar);

	int filter = pSliderBar->GetFilterMode();

	pSliderBar->SetNeutral();
	m_ColorMatrix = QGaussFilter::MatYUVtoRGB;
	m_Gamma = 1.0f;
	if (filter > 0) FilterOff();

	Invalidate(FALSE);
}

void CTinterView::OnFlipHorizontal()
{
	OnRotateFlip(RotateNoneFlipX);
}

void CTinterView::OnFlipVertical()
{
	OnRotateFlip(RotateNoneFlipY);
}

void CTinterView::OnRotate180()
{
	OnRotateFlip(Rotate180FlipNone);
}

void CTinterView::OnRotate270()
{
	OnRotateFlip(Rotate270FlipNone, true);
}

void CTinterView::OnRotate90()
{
	OnRotateFlip(Rotate90FlipNone, true);
}

void CTinterView::OnRotateFlip(RotateFlipType rft, bool bChangeSize)
{
	ASSERT(m_pOriginal);
	ASSERT(m_pModified);

	m_pOriginal->RotateFlip(rft);
	m_pModified->RotateFlip(rft);

	InvalidateDocument();

	if (bChangeSize)
	{
		SetSizes();
		FitInWindow();
	}
	Invalidate(bChangeSize);
}

void CTinterView::OnUpdateRotateFlip(CCmdUI * pCmdUI)
{
	pCmdUI->Enable(m_pModified != NULL && ! m_bGifPreview);
}

// Edit=>Size command

void CTinterView::OnEditSize()
{
	// Edit size also sets m_pOriginal to m_pModified
	CSizeDlg dlg(this);
	dlg.m_Width = m_Rect.Width;
	dlg.m_Height = m_Rect.Height;
	if (dlg.DoModal() == IDOK)
	{
		UINT w = dlg.m_Width;
		UINT h = dlg.m_Height;

		Bitmap * pNew = new Bitmap(w, h, PixelFormat24bppRGB);
		if (! pNew) return;

		if (pNew->GetLastStatus() == Ok)
		{
			QImageProperties propSource(m_pOriginal);
			QImageProperties propDest(pNew);
			propDest.Copy(propSource);

			Graphics g(pNew);
			g.SetInterpolationMode(InterpolationModeHighQualityBicubic);
			
			Rect rc(0, 0, w, h);
			ImageAttributes attr;

			attr.SetColorMatrix(& m_ColorMatrix);
			attr.SetGamma(m_Gamma);

			g.DrawImage(m_pModified,
				rc,
				0, 0,
				m_Rect.Width, m_Rect.Height,
				UnitPixel,
				& attr);

			SetBitmap(pNew);
			Invalidate();
			ZoomToPreset(ZOOM_100);
		}
		delete pNew;
	}
}

// View menu, zoom

void CTinterView::OnViewModified()
{
	m_bViewModified ^= true;
	if (m_pModified)
	{
		if (m_bGifPreview && ! m_bViewModified && m_pGif == NULL) CreateGif();
		UpdateBars();
		Invalidate(FALSE);
	}
}

void CTinterView::OnUpdateViewModified(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_bViewModified);
}

void CTinterView::OnViewBorder()
{
	m_bBorder ^= true;
	Invalidate(! m_bBorder);
}

void CTinterView::OnUpdateViewBorder(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_bBorder);
}

void CTinterView::OnViewZoomin()
{
	ZoomIn();
}

void CTinterView::OnUpdateViewZoomin(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pModified != NULL && CanZoomIn());
}

void CTinterView::OnViewZoomout()
{
	ZoomOut();
}

void CTinterView::OnUpdateViewZoomout(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pModified != NULL && CanZoomOut());
}

void CTinterView::OnViewZoomtowindow()
{
	ZoomToWindow();
}

void CTinterView::OnViewActualsize()
{
	ZoomToPreset(ZOOM_100);
}

void CTinterView::OnToolPlusloupe()
{
	SetZoomMode(ZoomViewZoomIn);
}

void CTinterView::OnUpdateToolPlusloupe(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(GetZoomMode() == ZoomViewZoomIn);
	OnUpdateGeneral(pCmdUI);
}

void CTinterView::OnToolMinusloupe()
{
	SetZoomMode(ZoomViewZoomOut);
}

void CTinterView::OnUpdateToolMinusloupe(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(GetZoomMode() == ZoomViewZoomOut);
	OnUpdateGeneral(pCmdUI);
}

void CTinterView::OnToolHand()
{
	SetZoomMode(ZoomViewDrag);
}

void CTinterView::OnUpdateToolHand(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(GetZoomMode() == ZoomViewDrag);
	OnUpdateGeneral(pCmdUI);
}

void CTinterView::OnZoom(UINT nID)
{
	ZoomToPreset(nID - ID_ZOOM_FIRST);
}

void CTinterView::OnUpdateZoom(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(pCmdUI->m_nID - ID_ZOOM_FIRST == (UINT) GetPresetZoom());
	OnUpdateGeneral(pCmdUI);
}

void CTinterView::OnZoomOther()
{
	CZoomLevelDlg dlg;
	dlg.m_ZoomPercentage = (int)(100.0f * GetZoom());

	if (dlg.DoModal() == IDOK)
	{
		ZoomTo((float) dlg.m_ZoomPercentage / 100.0f);

		CSize szDocument = GetTotalSize();
		ScrollToCenter(CPoint(szDocument.cx / 2, - szDocument.cy / 2));
	}
}

void CTinterView::OnViewBackgroundcolor()
{
	CColorDialog dlg(m_colBackGround, 0, this);
	if (dlg.DoModal() == IDOK)
	{
		m_colBackGround = dlg.GetColor();
		Invalidate();
	}
}

void CTinterView::OnViewBordercolor()
{
	CColorDialog dlg(m_colBorder, 0, this);
	if (dlg.DoModal() == IDOK)
	{
		m_colBorder = dlg.GetColor();
		if (m_bBorder && m_pModified) Invalidate();
	}
}

void CTinterView::OnViewProperties()
{
	if (! m_pModified) return;
	ASSERT(m_pOriginal);

	Bitmap * pImage = m_pOriginal;
	Bitmap * pFile = NULL;

	if (! GetDocument()->IsModified())
	{
		// If unmodified, get the properties of the file bitmap
		CString sPath = GetDocument()->GetPathName();
		if (sPath.GetLength()) pFile = new Bitmap(sPath.AllocSysString());
		if (! pFile) return;
		pImage = pFile;
	}

	QImagePropertiesDlg dlg(this);

	// Always store the properties in m_pOriginal
	if (dlg.Show(pImage, m_pOriginal, GetDocument()->GetPathName()))
		GetDocument()->SetModifiedFlag();

	delete pFile;
}

//===============
// TWAIN support

void CTinterView::OnFileSelectsource()
{
	SelectSource();
}

void CTinterView::OnUpdateFileSelectsource(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(IsValidDriver());	
}

void CTinterView::OnFileAcquire()
{
	Acquire();
}

void CTinterView::OnUpdateFileAcquire(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(SourceSelected());	
}

void CTinterView::CopyImage(HANDLE hBitmap, TW_IMAGEINFO& /*info*/)
{
	LPVOID p = ::GlobalLock(hBitmap);

	BITMAPINFO * pBMI = (BITMAPINFO *) p;
	Bitmap * pScanned = new Bitmap(pBMI, pBMI->bmiColors);

	if (pScanned && pScanned->GetLastStatus() == Ok)
	{
		SetBitmap(pScanned);
		FitInWindow();

		// mark as modified, because it still is unsaved.
		GetDocument()->SetModifiedFlag();
	}
	delete pScanned;

	::GlobalUnlock(p);
}

int CTinterView::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (QZoomView::OnCreate(lpCreateStruct) == -1) return -1;
	InitTwain(m_hWnd);
	return 0;
}

BOOL CTinterView::PreTranslateMessage(MSG* pMsg)
{
	if (ProcessMessage(*pMsg)) return TRUE;
	return QZoomView::PreTranslateMessage(pMsg);
}

void CTinterView::GetIdentity(void)
{
	CTwain::GetIdentity();

	m_AppId.Version.MajorNum = 1;
	m_AppId.Version.MinorNum = 3;
	strcpy (m_AppId.Version.Info, "1.3");
	strcpy (m_AppId.ProductName, "Tinter");
}

//==================
// handlers for status bar

void CTinterView::OnUpdateModified(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pModified && m_bViewModified);
}

void CTinterView::OnUpdateGifpreview(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pModified && m_bGifPreview && ! m_bViewModified);
}

void CTinterView::OnUpdateCalculating(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_Filter.IsCalculating());
}

void CTinterView::OnUpdateScale(CCmdUI *pCmdUI)
{
	CString s;
	s.Format(_T("%d%%"), Int(100.0f * GetZoom()));
	pCmdUI->SetText(s);
	pCmdUI->Enable(m_pModified != NULL);
}

void CTinterView::OnUpdateSize(CCmdUI *pCmdUI)
{
	CString s;
	s.Format(_T("%d \xd7 %d"), m_Rect.Width, m_Rect.Height);
	pCmdUI->SetText(s);
	pCmdUI->Enable(m_pModified != NULL);
}

//=================================
// GIF preview

void CTinterView::OnViewGifpreview()
{
	m_bGifPreview ^= true;

	if (m_bGifPreview && (m_pGif == NULL || GetDocument()->IsModified())) CreateGif();

	m_bViewModified = ! m_bGifPreview;
	UpdateBars();
	Invalidate(FALSE);
}

void CTinterView::OnUpdateViewGifpreview(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pModified != NULL);
	pCmdUI->SetCheck(m_bGifPreview ? 1 : 0);
}

void CTinterView::UpdateBars(void)
{
	CFrameWnd * pFrame = (CFrameWnd *) ::AfxGetMainWnd();
	CControlBar * pSliderBar = (CControlBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_SLIDERBAR);
	CDialogBar * pGifPreviewBar = (CDialogBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_GIFPREVIEWBAR);
	if (m_bGifPreview && ! m_bViewModified)
	{
		pFrame->ShowControlBar(pSliderBar, FALSE, FALSE);
		pFrame->ShowControlBar(pGifPreviewBar, TRUE, FALSE);
	}
	else
	{
		pFrame->ShowControlBar(pSliderBar, TRUE, FALSE);
		pFrame->ShowControlBar(pGifPreviewBar, FALSE, FALSE);
	}
}

void CTinterView::OnCloseGifPreview()
{
	m_bGifPreview = false;
	m_bViewModified = true;
	UpdateBars();
}

void CTinterView::OnGifMode(UINT nID)
{
	m_bGifOctree = (nID == IDC_OCTREE);
	CreateGif();
	Invalidate(FALSE);
}

void CTinterView::OnUpdateGifMode(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(pCmdUI->m_nID == (UINT)(m_bGifOctree ? IDC_OCTREE : IDC_WEBSAFE));
}

void CTinterView::CreateGif(void)
{
	delete m_pGif;
	m_pGif = NULL;

	// Bitmap m_pModified is in YUV color space. We need a clone in RGB space
	UINT w = m_pModified->GetWidth();
	UINT h = m_pModified->GetHeight();

	Bitmap * pSource = new Bitmap(w, h, PixelFormat24bppRGB);
	if (! pSource) return;

	if (pSource->GetLastStatus() == Ok)
	{
		CWaitCursor wait;

		{	// Block to scope g
			Graphics g(pSource);
			
			Rect rc(0, 0, w, h);
			ImageAttributes attr;
			attr.SetColorMatrix(& m_ColorMatrix);
			attr.SetGamma(m_Gamma);

			g.DrawImage(m_pModified,
				rc,
				0, 0,
				w, h,
				UnitPixel,
				& attr);
		}

		// Another pice de rsistance: our QColorQuantizer class
		QColorQuantizer quant;

		m_pGif = quant.GetQuantized(pSource, m_bGifOctree ? QColorQuantizer::Octree : QColorQuantizer::WebSafe);

		CGifPreviewBar * pGifPreviewBar = (CGifPreviewBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_GIFPREVIEWBAR);
		pGifPreviewBar->SetPalette(m_pGif);

		delete pSource;
	}
}

//=====================
// helpers

void CTinterView::InvalidateDocument(void)
{
	GetDocument()->SetModifiedFlag();
	delete m_pGif;
	m_pGif = NULL;
}

void CTinterView::SetBitmap(Bitmap * pSrc)
{
	if (m_bGifPreview) OnViewGifpreview();

	delete m_pOriginal;
	m_pOriginal = NULL;
	delete m_pModified;
	m_pModified = NULL;
	delete m_pGif;
	m_pGif = NULL;

	UINT w = pSrc->GetWidth();
	UINT h = pSrc->GetHeight();

	m_pOriginal = new Bitmap(w, h, PixelFormat24bppRGB);
	if (! m_pOriginal) return;

	if (m_pOriginal->GetLastStatus() == Ok)
	{
		Graphics g(m_pOriginal);
		
		Rect rc(0, 0, w, h);
		ImageAttributes attr;
		attr.SetColorMatrix(& QGaussFilter::MatRGBtoYUV);

		g.DrawImage(pSrc,
			rc,
			0, 0,
			w, h,
			UnitPixel,
			& attr);

		m_pModified = m_pOriginal->Clone(0, 0, w, h, PixelFormat24bppRGB);

		SetSizes();

		OnSetNeutral();
	}
}

void CTinterView::SetSizes(void)
{
	if (! m_pModified) return;

	UINT w = m_pModified->GetWidth();
	UINT h = m_pModified->GetHeight();

	m_szDoc.cx = w;
	m_szDoc.cy = h;

	m_Rect.Width = w;
	m_Rect.Height = h;

	int margin = ::GetSystemMetrics(SM_CXFRAME);

	m_szDoc.cx += 2 * margin;
	m_szDoc.cy += 2 * margin;

	m_Rect.X = margin;
	m_Rect.Y = margin;

	SetScrollSizes(MM_TEXT, m_szDoc);
}
